/*
 * Scratchpad MMU - Hardware driver for test program
 * Copyright (C) Jack Whitham 2009-2010
 * http://www.jwhitham.org.uk/c/smmu.html
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation
 * (version 2.1 of the License only).
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

#include <stdio.h>
#include <assert.h>
#include <string.h>
#include <stdlib.h>
#include "hwdriver.h"
#include "xparameters.h"

#define SUPPORTED_HW_VERSION    0x103

/* Control register map:
    0 - WR - initiate OPEN/CLOSE: program read only, size and copy_loc
        Rd - read status reg - result counter (hi)/result timer(lo)
    1 - WR - offset register - not required
        Rd - global counter - total nr of clock cycles since power on
    2 - WR - comment register - not required
        Rd - match counter - count SMMU table hits
    3 - WR - cache control register (if there is a cache)
        Rd - miss counter - count SMMU table misses
    4 - Rd - dbg_code from SMMU core
    5 - 
    6 - 
    7 - Rd - configuration register
*/


static unsigned         status_copy ;
static unsigned         DATA_MEMORY [ 1 ] ;

unsigned                NUM_RO_ENTRIES ;
unsigned                NUM_ENTRIES ;
unsigned                NUM_RW_ENTRIES ;
unsigned                FIRST_RW_ENTRY ;
unsigned                HAS_CACHE ;
unsigned                HAS_EXTRA_SP;

unsigned                LOG2_SP_SIZE ;
unsigned                LOG2_NUM_CACHE_WORDS_PER_LINE ;
unsigned                LOG2_NUM_CACHE_LINES_PER_PAGE ;
unsigned                LOG2_NUM_CACHE_PAGES ;

unsigned                SP_SIZE ;
unsigned                NUM_CACHE_WORDS_PER_LINE ;
unsigned                NUM_CACHE_LINES_PER_PAGE ;
unsigned                NUM_CACHE_PAGES ;

/* This function expects an address as a pointer,
 * a size in bytes, and a copy_loc in words */
unsigned HW_OPEN ( void * address , unsigned size , unsigned copy_loc )
{
#define ROF (1 << (31 - 16 + 2))
#define ROF31 (1 << 31)
    unsigned send , rc , error_code ;
    size = ( size + 3 ) >> 2 ;
    send = ( size << 16 ) | ( copy_loc & 0xffff ) ;

    assert (NUM_ENTRIES != 0);
    asm volatile (
            "swi    %3, %4, 0\n"        /* write to control */
            "lwi    %0, %5, 0\n"        /* read reference - do OPEN */
            "lwi    %1, %4, 0\n"
#ifdef TEST_MODE
            "lwi    %2, %4, 0x10\n"     /* get error code */
#endif
            : "=&r"(rc), "=&r"(status_copy), "=&r"(error_code)
            : "r"(send), "r"(CONTROL), "r"(address) );

#ifdef TEST_MODE
    if (( rc >= NUM_ENTRIES )
    || ( rc == 0 )
    || (( ! ( send & ROF31 )) && ( rc < FIRST_RW_ENTRY ))
    || (( send & ROF31 ) && ( rc >= FIRST_RW_ENTRY ))
    || (( error_code & 0xff ) != 0x09 ))
    {
        printf ( "\r\nOPEN failed with error code %08x "
            "and rc %08x status %08x command %08x\r\n",
            error_code, rc, status_copy, send) ;
        exit ( 1 ) ;
    }
#endif
    return rc ;
}

/* This function expects an address as a pointer,
 * a size in bytes, and a copy_loc in words */
unsigned HW_OPEN_RO ( void * address , unsigned size , unsigned copy_loc )
{
    return HW_OPEN ( address , size | ROF , copy_loc ) ;
}


void HW_CLOSE ( unsigned ref )
{
    unsigned error_code, initial_error_code;

    /*%0 - status copy
      %1 - error_code
      %2 - initial error code
      %3 - ref
      %4 - CONTROL
      %5 - DATA MEMORY
   */
    assert (NUM_ENTRIES != 0);
    asm volatile (
            "lwi    %2, %4, 0x10\n"
            "swi    r0, %4, 0\n"        /* write to control */
            "swi    %3, %5, 0\n"        /* write reference - do CLOSE */
            "lwi    %0, %4, 0\n"        /* get status */
#ifdef TEST_MODE
            "lwi    %1, %4, 0x10\n"     /* get error code */
#endif
            : "=&r"(status_copy), "=&r"(error_code), "=&r"(initial_error_code)
            : "r"(ref), "r"(CONTROL), "r"(DATA_MEMORY) ) ;

#ifdef TEST_MODE
    if ((( error_code & 0xff ) != 0xc9 )
    && ! (( ref != 0 ) && ( ref <= NUM_RO_ENTRIES ) 
        && (( error_code & 0xff ) == 0xc0 )))
    {
        printf ( "\r\nCLOSE failed with error code %08x while "
            "closing %d status %08x\r\n",
            error_code, ref, status_copy) ;
        exit ( 1 ) ;
    }
#endif
}

unsigned HW_Get_Result_Counter ( void )
{
    return status_copy ; //( status_copy >> 16 ) & 0xffff ;
}

unsigned HW_Get_Result_Timer ( void )
{
    return status_copy & 0xffff ;
}

void HW_Cache_Enable ( unsigned flush )
{
    unsigned control_shadow;

    if ( ! HAS_CACHE ) return;
    if ( flush ) 
    {
        unsigned d = NUM_CACHE_WORDS_PER_LINE * 4 ;
        unsigned e = SP_SIZE * 4 ;
        unsigned i;

        control_shadow = 0x2;
        for ( i = 0; i < e; i += d )
        {
            asm volatile ( 
                "swi    %0, %1, 0xc\n"  /* Write to control */
                "lw     r0, %2, %3\n"    /* This load causes the invalidate */
                : : "r"(control_shadow), "r"(CONTROL), "r"(i),
                    "r"(DATA_MEMORY) );
        }
    }
    control_shadow = 0x1;
    asm volatile ( 
        "swi    %0, %1, 0xc\n"  /* Write to control: activate cache */
        : : "r"(control_shadow), "r"(CONTROL) );
}

void HW_Cache_Disable ()
{
    asm volatile ( 
        "swi    r0, %0, 0xc\n"  /* Write to control: disable cache */
        : : "r"(CONTROL) );
}

void HW_Cache_Set_Permutation (unsigned src, unsigned dest)
{
    unsigned line_address = 
            src << (2 + LOG2_NUM_CACHE_WORDS_PER_LINE +
                    LOG2_NUM_CACHE_LINES_PER_PAGE);
    unsigned base_address = (unsigned) DATA_MEMORY;
    unsigned data = dest;
    unsigned control_shadow = 0x2;

    if ( ! HAS_CACHE ) return;
    base_address = 0x9 << 28; //&= ~(( 1 << LOG2_SP_SIZE ) - 1 );
    asm volatile ( 
        "swi    %3, %0, 0xc\n"  /* Write to control: disable cache */
        "sw     %2, %4, %1\n"   /* Must follow control write */
        : : "r"(CONTROL), "r"(line_address), "r"(data), "r"(control_shadow),
                "r"(base_address) );
}

int HW_Probe ( void )
{
    unsigned dbg_code = CONTROL->dbg_code;
    unsigned config = CONTROL->configuration;
    unsigned version = config >> 20;

    return ((version == SUPPORTED_HW_VERSION) 
        && ((dbg_code >> 24) == 0xdb));
}

void HW_Init ( void )
{
    unsigned dbg_code = CONTROL->dbg_code;
    unsigned config = CONTROL->configuration;
    unsigned version = config >> 20;
    unsigned REF_LEFT = (config >> 16) & 0xf;
    unsigned ACTIVE_LEFT = (config >> 12) & 0xf;
    unsigned SP_LEFT = (config >> 4) & 0xf;
    /* unsigned features = config & 0xf; */

    ACTIVE_LEFT += 0x10;
    NUM_ENTRIES = 0;
    if ((version == 0) || (version == 0xfff)
    || ((dbg_code >> 24) != 0xdb)
    || (ACTIVE_LEFT < 18) || (SP_LEFT < 2)
    || (REF_LEFT < 2)) {
        printf("SMMU initialisation failed. Hardware not detected.\r\n");
        printf("version %08x\n", config);
        printf("dbg_code %08x\n", dbg_code);
        exit(1);
    }
    if (version != SUPPORTED_HW_VERSION) {
        printf("SMMU initialisation failed.\r\nHardware version: %03x", version);
        printf("\r\nDriver version: %03x", SUPPORTED_HW_VERSION);
        printf("\r\nMismatched versions.\r\n");
        exit(1);
    }
    /* Attributes obtained direct from SMMU configuration */
    SP_SIZE = 1 << (SP_LEFT + 1);   /* in words */
    NUM_RO_ENTRIES = (config >> 8) & 0xf;
    NUM_ENTRIES = 1 << (REF_LEFT + 1);
    assert (NUM_ENTRIES > NUM_RO_ENTRIES);
    assert (NUM_ENTRIES <= MAX_NUM_ENTRIES);
    HAS_EXTRA_SP = !!(config & 0x8);

    /* Derived attributes */
    NUM_RW_ENTRIES = NUM_ENTRIES - (NUM_RO_ENTRIES + 1);
    FIRST_RW_ENTRY = NUM_RO_ENTRIES + 1;
    LOG2_SP_SIZE = SP_LEFT + 1;

    /* Cache setup (if there isn't a cache) */
    HAS_CACHE = 0;
    LOG2_NUM_CACHE_WORDS_PER_LINE = 0;
    LOG2_NUM_CACHE_LINES_PER_PAGE = 0;
    LOG2_NUM_CACHE_PAGES = 0;
    NUM_CACHE_WORDS_PER_LINE = 0;
    NUM_CACHE_LINES_PER_PAGE = 0;
    NUM_CACHE_PAGES = 0;

    /* Cache setup (if there is a cache) */
    if (config & 0x4) {
        unsigned i;
        unsigned cache_config = CONTROL->configuration_2;

        HAS_CACHE = 1;
        LOG2_NUM_CACHE_LINES_PER_PAGE = (cache_config >> 28) + 1;
        LOG2_NUM_CACHE_WORDS_PER_LINE = ((cache_config >> 24) & 0xf) + 1;
        LOG2_NUM_CACHE_PAGES = SP_LEFT + 1 -
            ( LOG2_NUM_CACHE_WORDS_PER_LINE + LOG2_NUM_CACHE_LINES_PER_PAGE );

        NUM_CACHE_WORDS_PER_LINE = 1 << LOG2_NUM_CACHE_WORDS_PER_LINE;
        NUM_CACHE_LINES_PER_PAGE = 1 << LOG2_NUM_CACHE_LINES_PER_PAGE;
        NUM_CACHE_PAGES = 1 << LOG2_NUM_CACHE_PAGES;

        assert (NUM_CACHE_WORDS_PER_LINE != 0);
        assert (NUM_CACHE_LINES_PER_PAGE != 0);
        assert (NUM_CACHE_PAGES != 0);
        assert (NUM_CACHE_PAGES <= SP_SIZE);

        for (i = 0; i < NUM_CACHE_PAGES; i++) {
            /* Default permutation - straight through */
            HW_Cache_Set_Permutation (i, i);
        }
    }
}

