/*
 * Scratchpad MMU - Simulator main module
 * Copyright (C) Jack Whitham 2009
 * http://www.jwhitham.org.uk/c/smmu.html
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation
 * (version 2.1 of the License only).
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */
#define QUIET

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <limits.h>
#include <stdint.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <signal.h>


#include "mb_mem.h"
#include "mb_core.h"
#include "mb_elf.h"
#include "smmu.h"
#include "base_ptrs.h"
#include "settings.h"

static int stop = 0 ;
static MB_Context * mb ;
#ifdef EARLY_TEST
static int deviant = 0 ;
static FILE * mcgrep_tape ;
static unsigned d_gpr [ 32 ] ;
static unsigned d2_gpr [ 32 ] ;
#endif

static void Trace ( void * t_user , MB_Context * mc , 
                            MB_Trace_Name trace_name , const void * param )
{
    unsigned param_u = (unsigned) ( (uintptr_t) param ) ;

#ifdef EARLY_TEST
    if (( trace_name == MB_EXECUTE )
    && ( ! deviant ))
    {
        /* Compare exec trace against the trace from
         * another simulator */
        if ( mcgrep_tape != NULL )
        {
            char buf [ BUFSIZ ] ;
            unsigned expect_pc , i ;

            buf [ 0 ] = '\0' ;
            if ( fgets ( buf , BUFSIZ , mcgrep_tape ) == NULL )
            {
                fclose ( mcgrep_tape ) ;
                mcgrep_tape = NULL ;
                return ;
            }
            buf [ BUFSIZ - 1 ] = '\0' ;
            expect_pc = (unsigned) strtoll ( buf , NULL , 16 ) ;
            for ( i = 1 ; i < 32 ; i ++ )
            {
                unsigned check = (unsigned) strtoll ( & buf [ ( i * 9 ) ] , 
                                NULL , 16 ) ;
                if (( check != 0xdeadbeef )
                && ( check != d_gpr [ i ] ))
                {
                    printf ( "Deviation with r%u at PC %x, is %x, "
                                "should be %x\n" ,
                                i , mc -> cur_pc , 
                                d_gpr [ i ] , check ) ;
                }
                d_gpr [ i ] = d2_gpr [ i ] ;
                d2_gpr [ i ] = mc -> gpr [ i ] ;
            }

            if ( expect_pc != mc -> cur_pc )
            {
                printf ( "Deviation at PC %x, should be %x\n" ,
                            mc -> cur_pc , expect_pc ) ;
                deviant = 1 ;
            }

        }
        if ( ! deviant )
        {
            return ;
        }
    }
#endif
    /* Exit command */
    if (( trace_name == MB_SIM_CMD )
    && ( param_u == 1 ))
    {
        stop = 1 ;
    }
    /* Operations that shouldn't be printed */
#ifdef QUIET
    if (( trace_name != MB_RESET )
    && ( trace_name != MB_DIV_BY_ZERO )
    && ( trace_name != MB_ILLEGAL_INST )
    && ( trace_name != MB_ALIGN ))
    {
        return ;
    }
#endif
    printf ( "%08x %08x %08x " , 
                    mc -> cur_pc , mc -> cur_iword , mc -> clock_cycle ) ;
    switch ( trace_name )
    {
        case MB_RESET :
                printf ( "MB_RESET" ) ;
                break ;
        case MB_INTERRUPT :
                printf ( "MB_INTERRUPT" ) ;
                break ;
        case MB_SPR_READ :
                printf ( "MB_SPR_READ" ) ;
                break ;
        case MB_SPR_WRITE :
                printf ( "MB_SPR_WRITE, value 0x%x" , param_u ) ;
                break ;
        case MB_CALL :
                printf ( "MB_CALL to pc 0x%x" , param_u ) ;
                break ;
        case MB_OPEN :
                printf ( "MB_OPEN %s" , (const char *) param ) ;
                break ;
        case MB_CLOSE :
                printf ( "MB_CLOSE %s" , (const char *) param ) ;
                break ;
        case MB_MISS_LOAD :
                printf ( "MB_MISS_LOAD 0x%x" , param_u ) ;
                break ;
        case MB_MISS_STORE :
                printf ( "MB_MISS_STORE 0x%x" , param_u ) ;
                break ;
        case MB_RETURN :
                printf ( "MB_RETURN to pc 0x%x" , param_u ) ;
                break ;
        case MB_ILLEGAL_INST :
                stop = 1 ;
                printf ( "MB_ILLEGAL_INST" ) ;
                break ;
        case MB_DIV_BY_ZERO :
                stop = 1 ;
                printf ( "MB_DIV_BY_ZERO" ) ;
                break ;
        case MB_EXECUTE :
                printf ( "MB_EXECUTE %s" , (const char *) param ) ;
                return ;
        case MB_SIM_CMD :
                printf ( "MB_SIM_CMD" ) ;
                break ;
        case MB_ALIGN :
                stop = 1 ;
                printf ( "MB_ALIGN" ) ;
                break ;
        default :
                //printf ( "Unknown trace name (%d)" , (int) trace_name ) ;
                break ;
    }
    printf ( "\n" ) ;
}

#define XUN_LSR_TX_BUFFER_EMPTY     0x20 /* Transmit holding reg empty */
#define XUN_THR_OFFSET  0x03 /* Transmit holding register */

static unsigned UART_Load ( void * io_user , unsigned address , unsigned size )
{
    return XUN_LSR_TX_BUFFER_EMPTY ; 
}

static void UART_Store ( void * io_user , unsigned address , 
                            unsigned data , unsigned size )
{
    address &= 0x1f ;
    if ( address == XUN_THR_OFFSET )
    {
        fputc ( (char) data , stdout ) ;
        fflush ( stdout ) ;
    }
}

static void Program_Aborted ( int signum, siginfo_t * siginfo, void * data )
{
    if ( mb != NULL )
    {
        printf ( "Aborted at PC %08x, instruction %08x, r15 = %08x\n" , 
                mb -> cur_pc , mb -> cur_iword,
                mb -> gpr [ 15 ] ) ;
    }
    exit ( 1 );
}


int main ( int argc , char ** argv )
{
    struct MB_System_Context_struct * sys ;
    struct SMMU_Context_struct *    sc ;
    const char *                    out ;

    if ( argc != 2 )
    {
        printf ( "Usage: %s <elf binary>\n" , argv [ 0 ] ) ;
        return 1 ;
    }
    sys = MB_System_Init ( Trace , EXT_MEM_LATENCY , NULL ) ;
    mb = MB_System_Get_MB_Context ( sys ) ;
    out = MB_Read_Elf ( sys , argv [ 1 ] ) ;
    if ( out != NULL )
    {
        printf ( "ELF read error: %s\n" , out ) ;
        return 1 ;
    }
    /* Register minimal support for the UART device */
    MB_Map_IO ( sys , 0x83e00000 >> PAGE_SHIFT , 
                    UART_Load , UART_Store , NULL ) ;
    /* Register SMMU */
    sc = SMMU_Init ( sys , NUM_RW_ENTRIES , NUM_RO_ENTRIES , SP_SIZE ) ;
    //SMMU_Cache_Init (sc, 3, 3);

    /* SIGABRT is raised by assertion failure */
    {
        struct sigaction sa;
        sa.sa_sigaction = Program_Aborted;
        sa.sa_flags = SA_SIGINFO;
        sigemptyset(&sa.sa_mask);
        sigaction(SIGABRT, &sa, NULL);
    }

    /* Simulate */
    while ( ! stop )
    {
        MB_Step ( mb , 0 ) ;
    }
    mb = NULL;
    SMMU_Delete ( sc ) ;
    MB_System_Delete ( sys ) ;
    return 0 ;
}


